@@ -1,13 +1,18 @@
 from weakref import WeakSet
 import numpy as np
 import colorlog
-from .utils import classproperty, get_obj_by_id
+from .utils import classproperty, get_obj_by_id, get_random_color
 
 from .config import cs_axis_up
-from .ParameterStructure.Instances import InstanceType
+
+from System.Windows.Media.Media3D import Point3D
+from System import Array
+from .GeometryViewer.Model.Geometry import Vertex, Edge, EdgeLoop, Face, Volume, Layer, PEdge, GeometricOrientation
+from .GeometryViewer.Model import DerivedColor
+from .ParameterStructure.Instances import SimInstanceType
 # from copy import deepcopy
 # from functools import lru_cache
-from tqdm import tqdm
+# from tqdm import tqdm
 from copy import copy
 
 logger = colorlog.getLogger('PySimultan')
@@ -297,12 +302,74 @@ class BaseGeoBaseClass(object):
 
 class BaseGeometricLayer(BaseGeoBaseClass):
 
+    @classmethod
+    def create_new(cls, geometry_model, name: str, layer=None, **kwargs):
+
+        l = Layer(geometry_model.Geometry, name)
+
+        if layer is None:
+            geometry_model.Geometry.Layers.Add(l)
+        else:
+            layer._wrapped_obj.Layers.Add(l)
+
+        layer = cls(geometry_model=geometry_model,
+                    wrapped_obj=l)
+
+        layer.color = DerivedColor(get_random_color())
+
+        for key, value in kwargs.items():
+            setattr(layer, key, value)
+
+        return layer
+
     def __init__(self, *args, **kwargs):
         BaseGeoBaseClass.__init__(self, *args, **kwargs)
 
+    @property
+    def color(self):
+        return self._wrapped_obj.Color
+
+    @color.setter
+    def color(self, value):
+        self._wrapped_obj.Color = value
+
+    @property
+    def is_visible(self):
+        return self._wrapped_obj.IsVisible
+
+    @is_visible.setter
+    def is_visible(self, value: bool):
+        self._wrapped_obj.IsVisible = value
+
 
 class BaseGeometricVertex(BaseGeoBaseClass):
 
+    @classmethod
+    def create_new(cls, geometry_model, position, layer=None, **kwargs):
+
+        if layer is None:
+            layer = geometry_model.layers[0]._wrapped_obj
+        if isinstance(layer, BaseGeometricLayer):
+            layer = layer._wrapped_obj
+
+        simultan_position = position
+        if cs_axis_up == 1:
+            simultan_position = np.array([position[0], position[1], position[2]])
+        elif cs_axis_up == 2:
+            simultan_position = np.array([-position[0], position[2], position[1]])
+        elif cs_axis_up == 3:
+            simultan_position = np.array([position[1], position[0], position[2]])
+
+        v = Vertex(layer, Point3D(simultan_position[0], simultan_position[1], simultan_position[2]))
+
+        vertex = cls(geometry_model=geometry_model,
+                     wrapped_obj=v)
+
+        for key, value in kwargs.items():
+            setattr(vertex, key, value)
+
+        return vertex
+
     def __init__(self, *args, **kwargs):
         BaseGeoBaseClass.__init__(self, *args, **kwargs)
 
@@ -335,6 +402,24 @@ class BaseGeometricVertex(BaseGeoBaseClass):
 
 class BaseGeometricEdge(BaseGeoBaseClass):
 
+    @classmethod
+    def create_new(cls, geometry_model, v1, v2, layer=None, **kwargs):
+
+        if layer is None:
+            layer = geometry_model.layers[0]._wrapped_obj
+        if isinstance(layer, BaseGeometricLayer):
+            layer = layer._wrapped_obj
+
+        e = Edge(layer, Array[Vertex]([v1._wrapped_obj, v2._wrapped_obj]))
+
+        edge = cls(geometry_model=geometry_model,
+                   wrapped_obj=e)
+
+        for key, value in kwargs.items():
+            setattr(edge, key, value)
+
+        return edge
+
     def __init__(self, *args, **kwargs):
         BaseGeoBaseClass.__init__(self, *args, **kwargs)
         self._vertices = kwargs.get('vertices', None)
@@ -348,6 +433,25 @@ class BaseGeometricEdge(BaseGeoBaseClass):
 
 class BaseGeometricEdgeLoop(BaseGeoBaseClass):
 
+    @classmethod
+    def create_new(cls, geometry_model, edges, layer=None, **kwargs):
+
+        if layer is None:
+            layer = geometry_model.layers[0]._wrapped_obj
+        if isinstance(layer, BaseGeometricLayer):
+            layer = layer._wrapped_obj
+
+        el = EdgeLoop(layer, Array[Edge]([x._wrapped_obj for x in edges]))
+
+        edge_loop = cls(geometry_model=geometry_model,
+                        wrapped_obj=el)
+
+        for key, value in kwargs.items():
+            setattr(edge_loop, key, value)
+
+        return cls(geometry_model=geometry_model,
+                   wrapped_obj=el)
+
     def __init__(self, *args, **kwargs):
         BaseGeoBaseClass.__init__(self, *args, **kwargs)
         self._edges = kwargs.get('edges', None)
@@ -383,6 +487,37 @@ class BaseGeometricEdgeLoop(BaseGeoBaseClass):
 
 class BaseGeometricFace(BaseGeoBaseClass):
 
+    @classmethod
+    def create_new(cls, geometry_model, edge_loop, orientation=None, openings=None, layer=None, **kwargs):
+
+        if layer is None:
+            layer = geometry_model.layers[0]._wrapped_obj
+        if isinstance(layer, BaseGeometricLayer):
+            layer = layer._wrapped_obj
+
+        if openings is None:
+            openings = []
+
+        if orientation is None:
+            orientation = GeometricOrientation.Forward
+
+        if orientation == 'forward':
+            orientation = GeometricOrientation.Forward
+        elif orientation == 'backward':
+            orientation = GeometricOrientation.Backward
+        else:
+            orientation = GeometricOrientation.Undefined
+
+        f = Face(layer, edge_loop._wrapped_obj, orientation, Array[EdgeLoop]([x._wrapped_obj for x in openings]))
+
+        new_obj = cls(geometry_model=geometry_model,
+                      wrapped_obj=f)
+
+        for key, value in kwargs.items():
+            setattr(new_obj, key, value)
+
+        return new_obj
+
     def __init__(self, *args, **kwargs):
         BaseGeoBaseClass.__init__(self, *args, **kwargs)
 
@@ -399,7 +534,7 @@ class BaseGeometricFace(BaseGeoBaseClass):
 
     @property
     def construction_assignment(self):
-        return next((x for x in self._components if x.InstanceType == InstanceType.ALIGNED_WITH), None)
+        return next((x for x in self._components if x.InstanceType == 4), None)
 
     @property
     def construction(self):
@@ -469,9 +604,42 @@ class BaseGeometricFace(BaseGeoBaseClass):
     def hull_face(self):
         return None in [self.side_1_volume, self.side_2_volume]
 
+    def associate_with_component(self, component, exch=None):
+        """
+
+        :param component: Instance of PySimultan.default_types.SimultanObject, PySimultan.default_types.List, PySimultan.default_types.ReferenceList
+        :param exch: SIMULTAN.DataExchange import ComponentGeometryExchange
+        """
+        try:
+            component.InstanceType = SimInstanceType.Attributes2D
+        except Exception as e:
+            raise e
+
+        if exch is None:
+            exch = self.geometry_model.template_parser.current_data_model.exch
+        exch.Associate(component, self._wrapped_obj)
+
 
 class BaseGeometricVolume(BaseGeoBaseClass):
 
+    @classmethod
+    def create_new(cls, geometry_model, faces, layer=None, **kwargs):
+
+        if layer is None:
+            layer = geometry_model.layers[0]._wrapped_obj
+        if isinstance(layer, BaseGeometricLayer):
+            layer = layer._wrapped_obj
+
+        v = Volume(layer, Array[Face]([x._wrapped_obj for x in faces]))
+
+        new_obj = cls(geometry_model=geometry_model,
+                      wrapped_obj=v)
+
+        for key, value in kwargs.items():
+            setattr(new_obj, key, value)
+
+        return new_obj
+
     def __init__(self, *args, **kwargs):
         BaseGeoBaseClass.__init__(self, *args, **kwargs)
         self._faces = kwargs.get('faces', None)
@@ -480,6 +648,15 @@ class BaseGeometricVolume(BaseGeoBaseClass):
     def components(self):
         return self.geometry_model.get_py_geo_components(self._wrapped_obj)
 
+    def associate_with_component(self, component):
+        try:
+            component.InstanceType = SimInstanceType.Entity3D
+        except Exception as e:
+            raise e
+
+        exch = self.geometry_model.template_parser.current_data_model.exch
+        exch.Associate(component, self._wrapped_obj)
+
 
 def create_geo_classes(geo_types):
     """
