@@ -3,6 +3,11 @@ from .config import yaml
 import io
 import colorlog
 from uuid import uuid4
+import typing
+
+from yamale.readers.yaml_reader import _parsers
+import yamale
+
 
 from collections import UserList
 from .default_types import SimultanObject
@@ -10,12 +15,27 @@ from .default_types import List as SimultanList
 from .default_types import ValueField, BuildInFace, BuildInVolume, BuildInZone, BaseBuildInConstruction, BuildInMaterialLayer, ReferenceList
 from .geo_default_types import geometry_types
 from .geometry import GeometryModel
+from .utils import parse_slot
 
 from functools import lru_cache
 
 from System.Linq import Enumerable
 from ParameterStructure.Values import SimMultiValuePointer
-from .ParameterStructure.Component import Component
+from .ParameterStructure.Components import SimComponent
+
+from typing import List as TypeList
+from typing import NewType, Union
+
+try:
+    import importlib.resources as pkg_resources
+except ImportError:
+    # Try backported to PY<37 importlib_resources.
+    import importlib_resources as pkg_resources
+
+from . import resources
+
+with pkg_resources.path(resources, 'schema.yaml') as r_path:
+    schema_file = str(r_path)
 
 
 logger = colorlog.getLogger('PySimultan')
@@ -28,24 +48,25 @@ template_classes = {}
 @yaml_object(yaml)
 class Template(object):
 
-    yaml_tag = u'!Template'
+    yaml_tag = u'!Template:'
 
     def __init__(self, *args, **kwargs):
-        """ Template class to define how SIMULTAN components are imported and parsed
+        """ Template class to define how SIMULTAN SimComponents are imported and parsed
 
         :param args:
         :param kwargs:
 
         @keyword template_name: Name of the Template. This name is the name of the created Python class and used for
-        TYPE matching the SIMULTAN component
+        TYPE matching the SIMULTAN SimComponent
         @keyword template_id: ID of the template; int
         @keyword inherits_from: name of the python template or the template instance this template inherits from
         @keyword content: List of attribute names
         @keyword documentation: Entry to document the template; str
         @keyword units: units of the content. Dictionary
         @keyword types: Dictionary with the type of the content. if the type is 'str' get_TextValue() is returned, param.get_ValueCurrent() otherwise
-        @keyword slots: Dictionary with the name of the content and the name of the slot where it is expected
-
+        @keyword slots: Dictionary with the name of the content and the name of the slot where it is expected. The expected slot is a string
+        @keyword default_slot: Default slot name for the Simultan component
+        with the slot name and the extension, seperated by a blank
         """
 
         self.template_name = kwargs.get('template_name', None)
@@ -64,10 +85,12 @@ class Template(object):
         self.template_parser = kwargs.get('template_parser', None)
         self.template_class = kwargs.get('template_class', None)
 
-    def write(self, filename=None):
+        self.default_slot = kwargs.get('default_slot', 'Element')
+
+    def write(self, filename: str = None):
         """
         write the template as .yml to a file
-        :param filename: name and path of the file to save
+        :param filename: name and path of the file to save; Example: C:\SIMULTAN\my_template.yml
         :return: None
         """
         if filename is not None:
@@ -138,8 +161,9 @@ class Template(object):
                           '_types': self.types,
                           '_units': self.units,
                           '_base': bases,
-                          '_template_parser': self,
-                          '_slots': self.slots}
+                          '_template_parser': self.template_parser,
+                          '_slots': self.slots,
+                          '_default_slot': self.default_slot}
 
         new_class_dict.update(self.get_properties())
         new_class = type(self.template_name, bases, new_class_dict)
@@ -154,9 +178,14 @@ class Template(object):
 
         for prop in self.content:
 
+            if prop in self.slots.keys():
+                slot = self.slots[prop]
+            else:
+                slot = None
+
             syn = self.synonyms.get('prop', prop)
 
-            prop_dict[syn] = add_properties(prop, type=self.types.get(prop, None))
+            prop_dict[syn] = add_properties(prop, param_type=self.types.get(prop, None), slot=slot)
 
         return prop_dict
 
@@ -211,13 +240,14 @@ class TemplateParser(object):
 
         """
 
+        self._current_data_model = None
+
         self.id = uuid4()
 
         self._templates = kwargs.get('templates', None)
         self.template_filepath = kwargs.get('template_filepath', None)
 
         self._template_classes = kwargs.get('template_classes', None)
-        self._current_data_model = kwargs.get('current_data_model', None)
 
         self.data_models = kwargs.get('data_models', {})
 
@@ -225,6 +255,8 @@ class TemplateParser(object):
 
         template_classes[self.id] = self._template_classes
 
+        self.current_data_model = kwargs.get('current_data_model', None)
+
     @property
     def current_data_model(self):
         return self._current_data_model
@@ -232,6 +264,9 @@ class TemplateParser(object):
     @current_data_model.setter
     def current_data_model(self, value):
         self._current_data_model = value
+        if self._current_data_model is None:
+            return
+
         if self._current_data_model not in self.data_models:
             self.data_models[self._current_data_model.id] = self._current_data_model
 
@@ -271,54 +306,62 @@ class TemplateParser(object):
         if filepath is None:
             filepath = self.template_filepath
 
-        with open(filepath, mode='r', encoding="utf-8") as f_obj:
-            templates = yaml.load(f_obj)
+        if filepath is None:
+            templates = []
+        else:
+            with open(filepath, mode='r', encoding="utf-8") as f_obj:
+                templates = yaml.load(f_obj)
 
         return templates
 
-    def create_template_classes(self):
-
-        template_classes = {**self.bases}
+    def create_template_classes(self) -> typing.Dict[str, str]:
 
         logger.info('\n\nCreating template-classes:\n-------------------------------------')
-        logger.info(f'found {self.templates.__len__()} templates')
+        template_classes = {**self.bases}
 
-        for template in self.templates:
+        if self.templates is not None:
+            logger.info(f'found {self.templates.__len__()} templates')
 
-            new_class = template.create_template_class(self, template_classes)
-            template_classes[template.template_name] = new_class
+            for template in self.templates:
 
-            logger.info(f'Created template class {template.template_name}. Inherits from {template.inherits_from}')
+                new_class = template.create_template_class(self, template_classes)
+                template_classes[template.template_name] = new_class
+
+                logger.info(f'Created template class {template.template_name}. Inherits from {template.inherits_from}')
+
+        else:
+            logger.info(f'No template file defined')
 
         self._template_classes = template_classes
         logger.info('template-class creation finished\n\n')
+        return self._template_classes
 
     @lru_cache(maxsize=None)
-    def create_python_object(self, component, template_name=None):
+    def create_python_object(self, sim_component, template_name=None):
 
         # get the template or slot
         # template_name = None
-        if hasattr(component, 'ContainedParameters'):
-            t_name = next((x.TextValue for x in component.ContainedParameters.Items if x.Name == 'TYPE'), None)
+        if hasattr(sim_component, 'Parameters'):
+            t_name = next((x.TextValue for x in sim_component.Parameters.Items if x.Name == 'TYPE'), None)
             if t_name is not None:
                 template_name = t_name
 
-        if (template_name is None) and (hasattr(component, 'get_CurrentSlot')):
-            template_name = component.FitsInSlots[0]
+        if (template_name is None) and (hasattr(sim_component, 'get_CurrentSlot')):
+            template_name = sim_component.CurrentSlot.Base
 
-        # if no template and no slot could be found return the plain component
+        # if no template and no slot could be found return the plain SimComponent
 
         if template_name is None:
-            if isinstance(component, SimMultiValuePointer):
+            if isinstance(sim_component, SimMultiValuePointer):
                 template_name = 'ValueField'
 
         if template_name is None:
             if self._create_all:
-                if hasattr(component, 'ContainedComponentsAsList'):
-                    _ = [self.create_python_object(x) for x in component.ContainedComponentsAsList]
-                if hasattr(component, 'ReferencedComponents'):
-                    _ = [self.create_python_object(x) for x in component.ReferencedComponents.Items]
-            return component
+                if hasattr(SimComponent, 'ContainedSimComponentsAsList'):
+                    _ = [self.create_python_object(x) for x in sim_component.Components]
+                if hasattr(SimComponent, 'ReferencedSimComponents'):
+                    _ = [self.create_python_object(x) for x in sim_component.ReferencedComponents.Items]
+            return sim_component
 
         if template_name not in self.template_classes.keys():
 
@@ -347,7 +390,7 @@ class TemplateParser(object):
 
         # init new instance
         template_class._template_parser = self
-        new_instance = template_class(wrapped_obj=component,
+        new_instance = template_class(wrapped_obj=sim_component,
                                       template_parser=self,
                                       data_model_id=self._current_data_model.id)
 
@@ -366,72 +409,181 @@ class TemplateParser(object):
             # models.append(GeometryModel(wrapped_obj=model))
         return typed_models_dict
 
+    def add_new_geometry_model(self, file_name: str, model_name: str = None):
+
+        model, resource = self.current_data_model.add_new_geometry_model(file_name, model_name, return_resource=True)
+
+        return self.add_typed_geo_model(model, resource.Key)
+
+    def add_typed_geo_model(self, model, key):
+        """
+        Add a geometry model as typed model
+        :param model: ParameterStructure.Assets.ContainedResourceFileEntry
+        :param key: resource key
+        """
+        if self._typed_geo_models is None:
+            self._typed_geo_models = {}
+
+        self._typed_geo_models[key] = GeometryModel(template_parser=self,
+                                                    wrapped_obj=model,
+                                                    geo_types=self.geo_bases)
+
+        return self._typed_geo_models[key]
+
     def get_geo_components(self, geometry):
-        return Enumerable.ToList[Component](self.current_data_model.exch.GetComponents(geometry))
+        return Enumerable.ToList[SimComponent](self.current_data_model.exch.GetComponents(geometry))
 
     def get_py_geo_components(self, geometry, template_name=None):
-        components = Enumerable.ToList[Component](self.current_data_model.exch.GetComponents(geometry))
+        components = Enumerable.ToList[SimComponent](self.current_data_model.exch.GetComponents(geometry))
         return [self.create_python_object(x, template_name=template_name) for x in components]
 
 
-def add_properties(prop, type):
+def add_properties(prop_name: str, param_type: str, slot=None) -> property:
     """
     create property for a class
-    :param prop: name of the property (str)
-    :param type: type of the property; if type == 'str': param.get_TextValue(); else: param.get_ValueCurrent() is returned
+    :param prop_name: name of the property (str)
+    :param param_type: type of the property; if type == 'str': param.get_TextValue(); else: param.get_ValueCurrent() is returned
+    :param slot: slot; default is None
     :return: property
     """
 
     @lru_cache()
     def getx(self):
-
         obj = None
 
-        idx = next((i for i, x in enumerate(self._wrapped_obj.ContainedParameters.Items) if x.Name == prop), None)
+        idx = next((i for i, x in enumerate(self._wrapped_obj.Parameters.Items) if x.Name == prop_name), None)
         if idx is not None:
-            param = self._wrapped_obj.ContainedParameters.Items[idx]
+            param = self._wrapped_obj.Parameters.Items[idx]
             obj = param.get_MultiValuePointer()
 
             if obj is None:
-                if type == 'str':
+                if param_type == 'str':
                     obj = param.get_TextValue()
-                    # obj = next((x.get_TextValue() for x in self._wrapped_obj.ContainedParameters.Items if x.Name == prop), None)
                 else:
                     obj = param.get_ValueCurrent()
-                    # obj = next((x.get_ValueCurrent() for x in self._wrapped_obj.ContainedParameters.Items if x.Name == prop), None)
 
         if obj is None:
-            slot = self._slots.get(prop, None)
-            if slot is not None:
-                obj = next((x for x in self._wrapped_obj.ContainedComponentsAsList if x.get_CurrentSlot() == slot), None)
+            full_slot_name = self._slots.get(prop_name, None)
+            slot_name, slot_extension = parse_slot(full_slot_name)
+
+            if full_slot_name is not None:
+                obj = next((x.Component for x in self._wrapped_obj.Components.Items
+                            if (x.Slot.SlotBase.Base == slot_name) & (x.Slot.SlotExtension == slot_extension)),
+                           None)
+
                 if obj is None:
-                    obj = next((x.Reference for x in self._wrapped_obj.ReferencedComponents.Items if x.ReferenceFunction.SlotFull == slot), None)
+                    obj = next((x.get_Reference() for x in self._wrapped_obj.ReferencedComponents.Items
+                                if ' '.join([str(x.ReferenceFunction.SlotBase),
+                                             x.ReferenceFunction.SlotExtension]) == full_slot_name),
+                               None)
 
         return self._template_parser.create_python_object(obj)
 
-    def setx(self, value):
+    def setx(self, value: Union[int, float, SimComponent, SimultanObject]):
         getx.cache_clear()
-        return next((x.set_ValueCurrent(value) for x in self._wrapped_obj.ContainedParameters.Items if x.Name == prop), None)
+        idx = next((i for i, x in enumerate(self._wrapped_obj.Parameters.Items) if x.Name == prop_name), None)
+
+        if isinstance(value, (SimComponent, SimultanObject)):
+            # set reference or subcomponent
+
+            if prop_name not in self._slots.keys():
+                raise KeyError(f'{prop_name} not defined as slot for class {self.__class__.__name__}')
+
+            full_slot_name = self._slots.get(prop_name, None)
+            slot_name, slot_extension = parse_slot(full_slot_name)
+
+            # check if slot exists:
+            idx = self.slot_components.index[
+                (self.slot_components.slot == slot_name) & (self.slot_components.slot_extension == slot_extension)]
+
+            if idx.__len__() == 1:
+                # if subcomponent:
+                if self.slot_components.iloc[idx].comp_type.values == 0:
+                    if isinstance(value, SimultanObject):
+                        pass
+
+                    ref = next((x for x in self._wrapped_obj.Components.Items
+                                if ((x.Slot.SlotBase.Base == slot_name) & (
+                                x.Slot.SlotExtension == slot_extension))), None)
+
+                    if isinstance(value, SimComponent):
+                        ref.Component = value
+                    elif isinstance(value, SimultanObject):
+                        ref.Component = value._wrapped_obj
+
+                # if reference:
+                elif self.slot_components.iloc[idx].comp_type.values == 1:
+                    ref = next((x for x in self._wrapped_obj.ReferencedComponents.Items
+                                if ((x.ReferenceFunction.SlotBase.Base == slot_name) & (
+                                x.ReferenceFunction.SlotExtension == slot_extension))), None)
+
+                    if isinstance(value, SimComponent):
+                        # delete reference
+                        self.remove_component_reference(ref.reference)
+                        # add the SimComponent as sub-component
+                        self.add_sub_component(component=value,
+                                               slot_name=slot_name,
+                                               slot_extension=slot_extension,
+                                               alter_slot=True)
+                    else:
+                        ref.Reference = value._wrapped_obj
+
+            # index does not exist
+            elif idx.__len__() == 0:
+                if isinstance(value, SimultanObject):
+                    self.add_component_reference(component=value._wrapped_obj,
+                                                 slot_name=slot_name,
+                                                 slot_extension=slot_extension)
+                elif isinstance(value, SimComponent):
+                    self.add_sub_component(component=value, slot_name=slot_name, slot_extension=slot_extension)
+
+        elif param_type == 'str':
+            #todo: implementation for referenced parameters
+            if self._wrapped_obj.Parameters.Items[2].GetReferencedParameter() != self._wrapped_obj.Parameters.Items[idx]:
+                raise NotImplementedError(f'{self.name}: Param {prop_name} is a referenced Parameter. Setting referenced parameters is not yet implemented')
+            self._wrapped_obj.Parameters.Items[idx].set_TextValue(value)
+        elif isinstance(value, (int, float)):
+            # todo: implementation for referenced parameters
+            if self._wrapped_obj.Parameters.Items[2].GetReferencedParameter() != self._wrapped_obj.Parameters.Items[idx]:
+                raise NotImplementedError(f'{self.name}: Param {prop_name} is a referenced Parameter. Setting referenced parameters is not yet implemented')
+            self._wrapped_obj.Parameters.Items[idx].set_ValueCurrent(float(value))
+        else:
+            raise TypeError(
+                f'{self.name}: Error setting parameter {prop_name} to {type(value)}. Supported types are only: int float str SimComponent SimultanObject.')
 
     def delx(self):
         logger.warning('delete method not implemented')
+        raise NotImplementedError
 
     return property(getx, setx, delx, f"automatic created property")
 
 
-# def new_getattr(self, attr):
-#     try:
-#         return object.__getattribute__(self, attr)
-#     except KeyError:
-#         wrapped = object.__getattribute__(self, '_wrapped_obj')
-#         if wrapped is not None:
-#             return object.__getattribute__(wrapped, attr)
-#         else:
-#             raise KeyError
-#
-#
-# def new_setattr(self, attr, value):
-#     if (attr in self.__dict__) or (attr in ['_wrapped_obj', '_contained_components', '_contained_parameters']):
-#         object.__setattr__(self, attr, value)
-#     else:
-#         object.__setattr__(self._wrapped_obj, attr, value)
+def _my_loader(f):
+    import yaml
+
+    description = []
+
+    def any_constructor(loader, tag_suffix, node):
+        if isinstance(node, yaml.MappingNode):
+            return loader.construct_mapping(node)
+        if isinstance(node, yaml.SequenceNode):
+            return loader.construct_sequence(node)
+        return loader.construct_scalar(node)
+
+    yaml.add_multi_constructor('', any_constructor, Loader=yaml.SafeLoader)
+    data = list(yaml.load_all(f, Loader=yaml.SafeLoader))
+    return data
+
+
+_parsers['_my_loader'] = _my_loader
+
+
+class TemplateValidator(object):
+
+    def __init__(self, *args, **kwargs):
+
+        self.schema = yamale.make_schema(schema_file, parser='ruamel')
+
+    def validate(self, yaml_file):
+        data = yamale.make_data(yaml_file, parser='_my_loader')
+        yamale.validate(self.schema, data)
